/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   bufio.c								*
*									*
*	Read, write, manage buffers.  Why not use stdio?  We want	*
*	abit more control than that offers us.  For instance, all	*
*	sockets are non-blocking, but we sometimes want to block.	*
*									*
************************************************************************/
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>

#include "xmx.h"

/************************************************************************
*									*
*   forward routine defs						*
*									*
************************************************************************/
static int	write_it();
static void	adjust();
static void	trim_high_water();

extern char *	malloc();
extern char *	calloc();

/************************************************************************
*									*
*   local constants							*
*									*
************************************************************************/
#define	N_RETRIES	10	/* number of times to retry EWOULDBLOCKs */
#define	RETRY_WAIT_SEC	1	/* timeout for each retry */
#define	RETRY_WAIT_USEC	0
#define MIN_BUF_SIZE	2048
#define MIN_READ_SIZE	1024	/* half MIN_BUF_SIZE to prevent mallocs */
#define HIGH_WATER	8192	/* trim buffers larger than this */

/************************************************************************
*									*
*   global data								*
*									*
************************************************************************/
static buffer_t *uwd;		/* vector of unwritten data buffers */

/************************************************************************
*									*
*   buf_init								*
*									*
*	Initialize buffered io module.					*
*									*
************************************************************************/
void
buf_init(fdts)
   int fdts;
{
   register i;

   if (uwd)
      for (i=0; i<fdts; i++) {
         if (uwd[i].n)
            buf_clear(&uwd[i], 0);
      }
   else
      if ((uwd = (buffer_t *)calloc((unsigned)fdts, sizeof(buffer_t))) == 0)
         quit(1, "calloc returned zero\n");
}

/************************************************************************
*									*
*   buf_setwfds								*
*									*
*	Set fd mask (for select(2)) to reflect connections that need	*
*	to be written.							*
*									*
************************************************************************/
void
buf_setwfds(maxfd, fdsp)
   int maxfd;
   fd_set *fdsp;
{
   register i;

   for (i=0; i<=maxfd; i++)
      if (uwd[i].n)
         FD_SET(i, fdsp);
}

/************************************************************************
*									*
*   buf_clear								*
*									*
*	Clear active data in buffer.					*
*									*
************************************************************************/
void
buf_clear(bp, n)
   register buffer_t *bp;
   register n;		/* clear n bytes of active data (all if zero) */
{
   register i, t;

   if (n) {
      t = bp->n - n;
      for (i=0; i<t; i++)
         bp->cp[i] = bp->cp[i+n];
#ifdef PAINFUL
warn("buf_clear: [%x] cleared %d bytes (saved %d) size %d max %d\n",bp,n,t,bp->sz,t+n);
#endif
      bp->n = t >= 0 ? t : 0;
   }
   else
#ifdef PAINFUL
{warn("buf_clear: [%x] cleared all (%d)\n", bp, bp->n);
#endif
      bp->n = 0;
#ifdef PAINFUL
}
#endif
}

/************************************************************************
*									*
*   buf_zero								*
*									*
*	Zero out the buffer - assumes all values are garbage.		*
*									*
************************************************************************/
void
buf_zero(bp)
   register buffer_t *bp;
{
#ifdef PAINFUL
warn("buf_zero: [%x]\n", bp);
#endif
   bp->n = 0;
   bp->sz = 0;
   bp->cp = (char *)0;
}

/************************************************************************
*									*
*   buf_dup								*
*									*
*	Make an exact copy of a buffer.  Both copies will point to the	*
*	same buffer block.						*
*									*
************************************************************************/
void
buf_dup(bp1, bp2)
   register buffer_t *bp1, *bp2;
{
#ifdef PAINFUL
warn("buf_dup: [%x] [%x]\n", bp1, bp2);
#endif
   bp2->n = bp1->n;
   bp2->sz = bp1->sz;
   bp2->cp = bp1->cp;
}

/************************************************************************
*									*
*   buf_read								*
*									*
*	Read n bytes of data from fd into the buffer bp.  The buffer	*
*	is expanded as needed.  If n is zero, attempt to fill the	*
*	buffer.  If block is non-zero, the read blocks.  If block is	*
*	zero, this routine returns immediately.  Returns the number of	*
*	bytes read, or -1 if a read error occurred.			*
*									*
************************************************************************/
int
buf_read(fd, bp, n, block)
   int fd;
   buffer_t *bp;
   unsigned n;
   int block;
{
   register tries = 0;
   register r, t;
   register char *cp;
   fd_set rfds;
   struct timeval tv;

#ifdef PAINFUL
warn("buf_read: {%d} [%x] %d %s\n", fd, bp, n, block?"BLOCK":"NOBLOCK");
#endif
   if (n == 0) {
      t = bp->sz - bp->n;
      n = t > MIN_READ_SIZE ? t : MIN_READ_SIZE;
   }
   if (block) {
      tv.tv_sec = RETRY_WAIT_SEC;
      tv.tv_usec = RETRY_WAIT_USEC;
   }
   if (n > bp->sz - bp->n)	/* adjust buffer size if necessary */
      adjust(bp, n);

   cp = bp->cp + bp->n;
   for (t=0; t < n; t+=r)
      if ((r = read(fd, cp+t, n-t)) < 0)
         if (errno == EWOULDBLOCK)
            if (block) {
               r = 0;
               FD_ZERO(&rfds);
               FD_SET(fd, &rfds);
               (void)select(fd+1, &rfds, (fd_set*)0, (fd_set*)0, &tv);
               if (++tries == N_RETRIES) {
                  warn("buf_read: fd %d blocked for %d seconds...\n", fd,
						RETRY_WAIT_SEC * N_RETRIES);
                  tries = 0;
               }
               continue;
            }
            else
               break;
         else
            return -1;
      else if (r == 0)
#ifdef PAINFUL
{
warn("buf_read: EOF!\n");
#endif
         break;		/* EOF */
#ifdef PAINFUL
}
#endif
      else if (tries)
         tries = 0;

#ifdef PAINFUL
warn("buf_read: added %d bytes to %d active = %d\n", t, bp->n, bp->n + t);
#endif
   bp->n += t;

   return t;
}

/************************************************************************
*									*
*   buf_flush								*
*									*
*	Attempt to write out the unwritten data buffer for the given	*
*	file descriptor.  Returns the number of bytes remaining in	*
*	in the buffer, or -1 if a write error occurred.			*
*									*
************************************************************************/
int
buf_flush(fd, block)
   int fd, block;
{
   register i, r, t;

   if (uwd[fd].n) {
      r = write_it(fd, uwd[fd].cp, (int)uwd[fd].n, block);
      if (r < 0)			/* error */
#ifdef PAINFUL
{warn("buf_flush: write_it returned %d\n", r);
#endif
         return -1;
#ifdef PAINFUL
}
#endif
      else if (r == uwd[fd].n)		/* buffer flushed */
#ifdef PAINFUL
{warn("buf_flush: {%d} uwd flushed\n", fd);
#endif
         uwd[fd].n = 0;
#ifdef PAINFUL
}
#endif
      else if (r) {			/* partially flushed */
#ifdef PAINFUL
warn("buf_flush: {%d} partial %d of %d\n", fd, r, uwd[fd].n);
#endif
         t = uwd[fd].n - r;
         for (i=0; i<t; i++)
            uwd[fd].cp[i] = uwd[fd].cp[i+r];
         uwd[fd].n -= r;
      }
      if (uwd[fd].n <= MIN_BUF_SIZE)
         trim_high_water(&uwd[fd]);
   }
   return uwd[fd].n;
}

/************************************************************************
*									*
*   buf_put								*
*									*
*	Copy n bytes of raw data into a buffer block.			*
*									*
************************************************************************/
void
buf_put(bp, cp, n)
   buffer_t *bp;
   char *cp;
   unsigned n;
{
#ifdef PAINFUL
warn("buf_put: [%x] adding %d\n", bp, n);
#endif
   if (bp->sz - bp->n < n)
      adjust(bp, n);

   bcopy(cp, bp->cp + bp->n, n);
   bp->n += n;
}

/************************************************************************
*									*
*   buf_write								*
*									*
*	Write up to n bytes of data from the buffer bp onto fd.  If	*
*	n is zero, all active data in buf is attempted to be written.	*
*	If "block" is non-zero, the write blocks (though it will time	*
*	out and return an error eventually).  If block is zero, this	*
*	routine returns immediately.  In that case, if less than n	*
*	bytes of data were written, the remainder are queued in a 	*
*	per fd buffer to be written next time this routine or		*
*	buf_flush is called.						*
*									*
************************************************************************/
int
buf_write(fd, bp, n, block)
   int fd;
   buffer_t *bp;
   unsigned n;
   int block;
{
   register i, r, t;

#ifdef PAINFUL
warn("buf_write: {%d} [%x] %d %s\n", fd, bp, n, block?"BLOCK":"NOBLOCK");
#endif
   if (n == 0)
#ifdef PAINFUL
{warn("buf_write: zero, so write %d\n", bp->n);
#endif
      n = bp->n;
#ifdef PAINFUL
}
#endif
   else if (n > bp->n)
      quit(1, "buf_write: sanity check - write %d block contains %d\n",n,bp->n);

   if (r = buf_flush(fd, block)) {	/* flush old stuff */
      if (r < 0)
         return -1;
      buf_put(&uwd[fd], bp->cp, n);		/* queue all of it */
      r = n;
   }
   else {
      if ((r = write_it(fd, bp->cp, n, block)) < 0)
         return -1;
      if (r < n)
         buf_put(&uwd[fd], bp->cp+r, n-r);	/* queue unwritten portion */
   }

   if (n < bp->n) {	/* slide unwritten data to front of buffer block */
      t = bp->n - n;
#ifdef PAINFUL
warn("buf_write: sliding %d bytes forward\n", t);
#endif
      for (i=0; i<t; i++)
         bp->cp[i] = bp->cp[i+n];
   }
   bp->n -= n;
   if (bp->n <= MIN_BUF_SIZE)
      trim_high_water(bp);

   return n;
}

/************************************************************************
*									*
*   write_it								*
*									*
*	Generic write to non-blocking fd.  Repeatedly attempt to write	*
*	all data given until error or EWOULDBLOCK.  If block is non-	*
*	zero, block.							*
*									*
*	Return value is analogous to write system call.			*
*									*
************************************************************************/

static int
write_it(fd, cp, n, block)
   int fd;
   char *cp;
   int n, block;
{
   register tries = 0;
   register r, t;
   fd_set wfds;
   struct timeval tv;

#ifdef PAINFUL
warn("write_it: {%d} %d %s\n", fd, n, block?"BLOCK":"NOBLOCK");
#endif
   if (block) {
      tv.tv_sec = RETRY_WAIT_SEC;
      tv.tv_usec = RETRY_WAIT_USEC;
   }
   for (t=0; t < n; t+=r)
      if ((r = write(fd, cp+t, n-t)) < 0)
         if (errno == EWOULDBLOCK)
            if (block) {
               r = 0;
               FD_ZERO(&wfds);
               FD_SET(fd, &wfds);
               (void)select(fd+1, (fd_set*)0, &wfds, (fd_set*)0, &tv);
               if (++tries == N_RETRIES) {
                  warn("write_it: fd %d blocked for %d seconds...\n", fd,
						RETRY_WAIT_SEC * N_RETRIES);
                  tries = 0;
               }
               continue;
            }
            else
               break;
         else
            return -1;
      else
#ifdef PAINFUL
{warn("write_it: write wrote %d\n", r);
#endif
         if (tries) tries = 0;
#ifdef PAINFUL
}
#endif

#ifdef PAINFUL
warn("write_it: total written %d\n", t);
#endif
   return t;
}

/************************************************************************
*									*
*   adjust								*
*									*
*	Reallocate buffer block so that it has at least n bytes of	*
*	free space.							*
*									*
************************************************************************/
static void
adjust(bp, n)
   buffer_t *bp;
   register unsigned n;
{
   register sz;

#ifdef PAINFUL
warn("adjust: [%x] adjusting %d to %d\n", bp, bp->sz, bp->n + n);
#endif
   sz = bp->n + n;	/* note: can be smaller than bp->sz */

   bp->sz = sz > MIN_BUF_SIZE ? sz : MIN_BUF_SIZE;

   if (bp->cp) {
      if ((bp->cp = (char *)realloc(bp->cp, bp->sz)) == 0)
         quit(1, "realloc returned zero\n");
   }
   else
      if ((bp->cp = (char *)malloc(bp->sz)) == 0)
         quit(1, "malloc returned zero\n");
}

/************************************************************************
*									*
*	trim_high_water							*
*									*
*	Trim a buffer that has grown larger than an arbitrary		*
*	constant, but only after it hasn't been needed for at least	*
*	one write operation.						*
*									*
************************************************************************/
static void
trim_high_water(bp)
   buffer_t *bp;
{
return;	/* this code is ill-conceived and results in lost data */
/*NOTREACHED*/

   bp->hw <<= 2;	/* hw is a short, so we remember the last 8 calls */
   if (bp->sz > HIGH_WATER) {
      if (bp->hw == 0)
#ifdef PAINFUL
{warn("trim_high_water: trimming (%d)\n", bp->sz);
#endif
         adjust(bp, MIN_BUF_SIZE);
#ifdef PAINFUL
}
#endif
      bp->hw |= 3;
   }
}
