/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   connect.c								*
*									*
*       Create/check client/server connection blocks.			*
*									*
************************************************************************/
#include <stdio.h>
#include <X11/X.h>
#define NEED_REPLIES
#include <X11/Xproto.h>

#include "xmx.h"

			/* types for structifying a connection block */
typedef struct {
   xDepth		*d;
   xVisualType		*v;
}depth_t;

typedef struct {
   xWindowRoot		*r;
   depth_t		*d;
}root_t;

typedef struct {
   xConnSetupPrefix	*p;
   xConnSetup		*s;
   char 		*v;
   xPixmapFormat	*f;
   root_t		*r;
}setup_t;

static setup_t	*dsu=0;		/* default setup block */
static int	setup_len;	/* size of setup block (bytes) */
static buffer_t	dbuf;		/* default buffer: points to same setup block*/

static void	master_config();
static int	cmp_config();
static void	setup_print();
static setup_t	*structize();
static void	unstructize();

/************************************************************************
*									*
*   open_display							*
*									*
*	Just what it says: connect to server, send a client block	*
*	and read and test a setup block.  The block is saved and	*
*	returned in buffer, if supplied.  Returns the socket.		*
*	If an error occurred, a negative value is returned after	*
*	printing an error message.					*
*									*
************************************************************************/
int
open_display(hostname, display, buffer, bfactor)
   char *hostname;
   int display;
   buffer_t *buffer;	/* optional return value, block read */
   int bfactor;
{
   register s;
   register char *cp;
   int one = 1;
   static inited = 0;
   static xConnClientPrefix client;
   xConnSetupPrefix *prefix;
   static buffer_t buf;
   setup_t *sp;

   if (!inited) {
      client.byteOrder = *(char *)&one ? 'l' : 'B';
      client.majorVersion = X_PROTOCOL;
      client.minorVersion = X_PROTOCOL_REVISION;
      client.nbytesAuthProto = 0;
      client.nbytesAuthString = 0;

      buf_zero(&buf);
      inited = 1;
   }
   buf_clear(&buf, 0);
   buf_put(&buf, (char *)&client, sizeof(xConnClientPrefix));

			/* physically connect to server */
   if ((s = client_socket(hostname, display, bfactor)) < 0)
      return -1;
			/* send client block */
   if (buf_write(s, &buf, 0, BLOCK) < 0)
      return perr(-1, hostname);

			/* look for setup prefix */
   if (buf_read(s, &buf, sz_xConnSetupPrefix, BLOCK) < 0)
      return perr(-1, hostname);

   prefix = (xConnSetupPrefix *)buf.cp;
   if (prefix->success != xTrue)
      setup_len = prefix->lengthReason;
   else
      setup_len = prefix->length * 4;

			/* read the rest of setup */
   if (buf_read(s, &buf, setup_len, BLOCK) < 0)
      return perr(-1, hostname);

   prefix = (xConnSetupPrefix *)buf.cp;	/* cp may have been realloc'd */

			/* if connection not accepted, print message */
   if (prefix->success != xTrue) {
      if (setup_len)
         cp = (char *)(prefix+1);
      else
         cp = "No reason given";
      return err(-1, "Connection not accepted:\n\t[%s:%d] %s\n",
						hostname, display, cp);
   }
#ifdef DEBUG
   if (debug) {
      fprintf(stderr, "*******************************************\n");
      setup_print(prefix);
      fprintf(stderr, "*******************************************\n");
   }
#endif
   sp = structize((char *)prefix);/* turn the setup block into a structure */

				/* make sure it's compatible with master */
   if (dsu) {
      if (cmp_config(s, sp))
         return err(-1, "[%s:%d] slave config doesn't match master\n",
							hostname, display);
      unstructize(sp);			/* free block */
   }
   if (buffer)			/* return copy of buffer */
      buf_put(buffer, buf.cp, buf.n);

   if (!dsu) {			/* else save if this is the master */
      dsu = sp;
      buf_dup(&buf, &dbuf);
      buf_zero(&buf);
      master_config(s, dsu);
   }

   return s;
}

/************************************************************************
*									*
*   rootwindow								*
*									*
*	Returns the master server's root window id for the given	*
*	screen.								*
*									*
************************************************************************/
rid_t
rootwindow(screen)
{
   if (dsu)
      return dsu->r[screen].r->windowId;
   else
      return 0;
}

/************************************************************************
*									*
*   cmap_size								*
*									*
*	Returns the number of entries in the colormap for the given	*
*	visual.								*
*									*
************************************************************************/
int
cmap_size(visualid)
   register rid_t visualid;
{
   register i, j, k;

   for (i=0; i<dsu->s->numRoots; i++)
      for (j=0; j<dsu->r[i].r->nDepths; j++)
         for (k=0; k<dsu->r[i].d[j].d->nVisuals; k++)
            if (dsu->r[i].d[j].v[k].visualID == visualid)
               return dsu->r[i].d[j].v[k].colormapEntries;
   return 0;
}

/************************************************************************
*									*
*   test_client_block							*
*									*
*	Read a client block from socket s, test its contents and	*
*	return a boolean status (zero: success, non-zero: failure).	*
*									*
************************************************************************/
int
test_client_block(s)
   int s;
{
   int one = 1;
   static buffer_t buf;
   register xConnClientPrefix *client;

   buf_clear(&buf, 0);
   if (buf_read(s, &buf, sizeof(xConnClientPrefix), BLOCK) < 0)
      return perr(-1, "read client block");
   client = (xConnClientPrefix *)buf.cp;
   if (client->byteOrder != (*(char *)&one ? 'l' : 'B'))
      return err(-1, "client byte order differs - not yet supported\n");
   if (client->majorVersion > X_PROTOCOL)
      warn("warning: client talks a newer protocol\n");
   if (client->minorVersion != X_PROTOCOL_REVISION)
      warn("warning: client protocol revision does not match\n");

   return 0;
}

/************************************************************************
*									*
*   send_setup								*
*									*
*	Write buffer, or the default setup block if buffer is null, to	*
*	socket s.  A buffer should either always or never be supplied.	*
*	Conflicts may arise if the two modes are intermixed.		*
*									*
************************************************************************/
int
send_setup(s, buf)
   int s;
   buffer_t *buf;
{
   static unsigned long base = 0;	/* shifted left of the 20th bit */

   if (buf == 0) {
      dsu->s->ridMask = 0x7ffff;		/* low 19 bits (like Xsun) */
      dsu->s->ridBase = ++base << 20;	/* uniquely identify connection */
      buf = &dbuf;
      buf->n = setup_len + sz_xConnSetupPrefix;
   }
   if (buf_write(s, buf, 0, BLOCK) < 0)
      return perr(-1, "sending setup block");

   return 0;
}

/************************************************************************
*									*
*   master_config							*
*									*
*	Store the master servers config in rid map and resource model.	*
*									*
************************************************************************/
static void
master_config(s, sp)
   register s;
   setup_t *sp;
{
   register i, j, k;
   register ncols;

   map_server_init(s, sp->s->ridBase, sp->s->ridMask);
   for (i=0; i<sp->s->numRoots; i++) {
      for (j=0; j<sp->r[i].r->nDepths; j++)		/* fix this up... */
         for (k=0; k<sp->r[i].d[j].d->nVisuals; k++)
            if (sp->r[i].d[j].v[k].visualID == sp->r[i].r->rootVisualID)
               ncols = sp->r[i].d[j].v[k].colormapEntries;
      resource_new_root(	s,
				sp->r[i].r->windowId,
				sp->r[i].r->rootVisualID,
				sp->r[i].r->defaultColormap,
				ncols,
				sp->r[i].r->whitePixel,
				sp->r[i].r->blackPixel);
      for (j=0; j<sp->r[i].r->nDepths; j++)
         for (k=0; k<sp->r[i].d[j].d->nVisuals; k++)
            map_server_rid_add(s, sp->r[i].d[j].v[k].visualID, (char *)0, 0);
   }
}

/************************************************************************
*									*
*   cmp_config								*
*									*
*	Compare a setup structure with the saved (master)		*
*	configuration.  If the server support is too different,		*
*	returns a non-zero value.  If the server is close enough,	*
*	returns zero.							*
*									*
*	TODO: should check that depth formats match.			*
*									*
************************************************************************/
static int
cmp_config(s, sp)
   register s;
   setup_t *sp;
{
   register h, i, j, k, l, m, n;
   rid_t rids[512];

   if (sp->s->numRoots < dsu->s->numRoots)
      return err(-1, "slave server has too few screens\n");

		/* these are not hard errors because they aren't fatal */
   if (sp->s->imageByteOrder != dsu->s->imageByteOrder)
      warn("slave server image byte order differs\n");

   if (sp->s->bitmapBitOrder != dsu->s->bitmapBitOrder)
      warn("slave server bitmap bit order differs\n");

   if (sp->s->bitmapScanlineUnit != dsu->s->bitmapScanlineUnit)
      warn("slave server bitmap scan line unit differs\n");

   if (sp->s->bitmapScanlinePad != dsu->s->bitmapScanlinePad)
      warn("slave server bitmap scan line padding differs\n");

   /*
	The following code attempts to match the resources of the
	new display to those of the master.  In general, the new one
	can have more, but not fewer, resources than the master.  It
	must have at least as many screens.  Each screen must have at
	least as many - and the same - depths.  Finally each depth
	must have at least as many - and the same - visuals.

	We assume that the each screen has at least one depth.  We
	also assume that depth formats match.
   */
   h = n = 0;
				/* find a match for each master screen */
   for (i=0; i<dsu->s->numRoots; i++) {
      for (j=0; j<sp->s->numRoots; j++) {
         if (dsu->r[i].r->nDepths > sp->r[j].r->nDepths)
            continue;
				/* find a match for each screen's depth */
         for (k=0; k<dsu->r[i].r->nDepths; k++) {
            for (l=0; l<sp->r[j].r->nDepths; l++) {
               if (dsu->r[i].d[k].d->depth != sp->r[j].d[l].d->depth)
                  continue;
               if (dsu->r[i].d[k].d->nVisuals > sp->r[j].d[l].d->nVisuals)
                  continue;
               if (dsu->r[i].d[k].d->nVisuals == 0) {
                  if (k == 0) {
                     rids[h++] = sp->r[j].r->windowId;
                     rids[h++] = sp->r[j].r->defaultColormap;
                  }
                  n = -1;
                  break;
               }
				/* find a match for each depth's visual */
               for (m=0; m<dsu->r[i].d[k].d->nVisuals; m++) {
                  for (n=0; n<sp->r[j].d[l].d->nVisuals; n++) {
                     if (dsu->r[i].d[k].v[m].class != sp->r[j].d[l].v[n].class)
                        continue;
                     if (dsu->r[i].d[k].v[m].colormapEntries >
					sp->r[j].d[l].v[n].colormapEntries)
                        continue;
                     if (k == 0 && m == 0) {
                        rids[h++] = sp->r[j].r->windowId;
                        rids[h++] = sp->r[j].r->defaultColormap;
                     }
                     rids[h++] = sp->r[j].d[l].v[n].visualID;
                     n = -1;		/* Found matching Visual */
                     break;
                  }
                  if (n >= 0) break;	/* Give up on this Depth */
               }
               if (n < 0) break;	/* Found all matching Depths */
            }
            if (n >= 0) break;		/* Give up on this Screen */
         }
         if (n < 0) break;		/* Found all matching Screens */
      }
      if (n >= 0)			/* Give up on Display */
         return err(-1, "can't map slave server's configuration to master's\n");
   }
   map_server_init(s, sp->s->ridBase, sp->s->ridMask);
   for (i=0; i<h; i++)
      map_server_rid_add(s, rids[i], (char *)0);

   return 0;
}

/************************************************************************
*									*
*   setup_print								*
*									*
*	Spit out a server connection block to stderr.			*
*									*
************************************************************************/
static void
setup_print(prefix)
   xConnSetupPrefix *prefix;
{
   register i, j, k;
   char *cp;
   xConnSetup *setup;
   xPixmapFormat *format;
   xWindowRoot *screen;
   xDepth *depth;
   xVisualType *visual;
   char buf[1024];

   fprintf(stderr,"success:         %d\n", prefix->success);
   fprintf(stderr,"major.minor:     %d.%d\n",prefix->majorVersion,
						prefix->minorVersion);
   if (!prefix->success) {
      strncpy(buf, (char *)(prefix+1), prefix->lengthReason);
      buf[prefix->lengthReason] = '\0';
      fprintf(stderr,"reason: %s\n", buf);
   }
   else {
      setup = (xConnSetup *)(prefix + 1);
      fprintf(stderr,"release:         %d\n", setup->release);
      fprintf(stderr,"base.mask:       %x.%x\n",setup->ridBase, setup->ridMask);
      fprintf(stderr,"motion bufsize:  %d\n", setup->motionBufferSize);
      fprintf(stderr,"max req size:    %d\n", setup->maxRequestSize);
      fprintf(stderr,"num roots:       %d\n", setup->numRoots);
      fprintf(stderr,"num formats:     %d\n", setup->numFormats);
      fprintf(stderr,"byte order:      %s\n",
					setup->imageByteOrder?"MSBF":"LSBF");
      fprintf(stderr,"bit order:       %s\n",
					setup->bitmapBitOrder?"MSBF":"LSBF");
      fprintf(stderr,"scan unit.pad:   %d.%d\n", setup->bitmapScanlineUnit,
						setup->bitmapScanlinePad);
      fprintf(stderr,"keycode min.max: %d.%d\n", setup->minKeyCode,
						setup->maxKeyCode);

      cp = (char *)(setup + 1);
      i = setup->nbytesVendor;
      strncpy(buf, cp, i);
      buf[i] = '\0';
      fprintf(stderr,"vender:          %s\n", buf);

      format = (xPixmapFormat *)(cp + i + (4 - i % 4) % 4);
      for (i=setup->numFormats; i--; format++) {
         fprintf(stderr,"(format) depth:          %d\n", format->depth);
         fprintf(stderr,"         bits per pixel: %d\n", format->bitsPerPixel);
         fprintf(stderr,"         scan line pad:  %d\n", format->scanLinePad);
      }

      screen = (xWindowRoot *)format;
      for (i=setup->numRoots; i--;) {
         fprintf(stderr,"(screen) root:           %x\n", screen->windowId);
         fprintf(stderr,"         default cmap:   %x\n",
						screen->defaultColormap);
         fprintf(stderr,"         white.black:    %x.%x\n", screen->whitePixel,
						screen->blackPixel);
         fprintf(stderr,"         cur input mask: %x\n",
						screen->currentInputMask);
         fprintf(stderr,"         pix width.hite: %d.%d\n", screen->pixWidth,
						screen->pixHeight);
         fprintf(stderr,"         mm width.hite:  %d.%d\n", screen->mmWidth,
						screen->mmHeight);
         fprintf(stderr,"         min.max maps:   %d.%d\n",
						screen->minInstalledMaps,
						screen->maxInstalledMaps);
         fprintf(stderr,"         root visual:    %x\n", screen->rootVisualID);
         fprintf(stderr,"         backing store:  ");
         switch (screen->backingStore) {
            case 0:  fprintf(stderr,"Never\n");				break;
            case 1:  fprintf(stderr,"WhenMapped\n");			break;
            case 2:  fprintf(stderr,"Always\n");			break;
            default: fprintf(stderr,"%d\n", screen->backingStore);	break;
         }
         fprintf(stderr,"         save unders:    %d\n", screen->saveUnders);
         fprintf(stderr,"         root depth:     %d\n", screen->rootDepth);
         fprintf(stderr,"         num depths:     %d\n", screen->nDepths);
         depth = (xDepth *)(screen + 1);
         for (j=screen->nDepths; j--;) {
            fprintf(stderr,"(depth)  depth:       %d\n", depth->depth);
            fprintf(stderr,"         num visuals: %d\n", depth->nVisuals);
            visual = (xVisualType *)(depth + 1);
            for (k=depth->nVisuals; k--; visual++) {
               fprintf(stderr,"(visual) visual id:   %x\n", visual->visualID);
               fprintf(stderr,"         class:       ");
               switch (visual->class) {
                  case 0:  fprintf(stderr,"StaticGray\n");		break;
                  case 1:  fprintf(stderr,"GrayScale\n");		break;
                  case 2:  fprintf(stderr,"StaticColor\n");		break;
                  case 3:  fprintf(stderr,"PseudoColor\n");		break;
                  case 4:  fprintf(stderr,"TrueColor\n");		break;
                  case 5:  fprintf(stderr,"DirectColor\n");		break;
                  default: fprintf(stderr,"%d\n", visual->class);	break;
               }
               fprintf(stderr,"         bits/rgb:    %d\n", visual->bitsPerRGB);
               fprintf(stderr,"         cmap ents:   %d\n",
						visual->colormapEntries);
               fprintf(stderr,"         masks r.g.b: %x.%x.%x\n",
							visual->redMask,
							visual->greenMask,
							visual->blueMask);
            }
            depth = (xDepth *)visual;
         }
         screen = (xWindowRoot *)depth;
      }
   }
}

/************************************************************************
*									*
*   structize								*
*									*
*	make a server response block (xConnSetup, etc) randomly		*
*	accessible.							*
*									*
************************************************************************/
static setup_t *
structize(cp)
   char *cp;
{
   register i, j;
   setup_t *rv;
   xWindowRoot *rp;
   xDepth *dp;

   rv = (setup_t *)malloc(sizeof(setup_t));
   rv->p = (xConnSetupPrefix *)cp;
   rv->s = (xConnSetup *)(rv->p + 1);
   rv->v = (char *)(rv->s + 1);
   i = rv->s->nbytesVendor;
   rv->f = (xPixmapFormat *)(rv->v + i + (4 - i % 4) % 4);
   rv->r = (root_t *)malloc(sizeof(root_t) * rv->s->numRoots);
   rp = (xWindowRoot *)(rv->f + rv->s->numFormats);
   for (i=0; i<rv->s->numRoots; i++) {
      rv->r[i].r = rp;
      rv->r[i].d = (depth_t *)malloc(sizeof(depth_t)*rv->r[i].r->nDepths);
      dp = (xDepth *)(rv->r[i].r + 1);
      for (j=0; j<rv->r[i].r->nDepths; j++) {
         rv->r[i].d[j].d = dp;
         rv->r[i].d[j].v = (xVisualType *)(rv->r[i].d[j].d + 1);
         dp = (xDepth *)(rv->r[i].d[j].v + rv->r[i].d[j].d->nVisuals);
      }
      rp = (xWindowRoot *)dp;
   }
   return rv;
}

/************************************************************************
*									*
*   unstructize								*
*									*
*	free malloc'd space.						*
*									*
************************************************************************/
static void
unstructize(s)
   setup_t *s;
{
   register i;

   for (i=0; i<s->s->numRoots; i++)
      free((char *)s->r[i].d);
   free((char *)s->r);
   free((char *)s);
}
