/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   genreq.c								*
*									*
*	Routines that collect and broadcast X requests.			*
*									*
*	Note: we allow buf to grow without bound, assuming that the	*
*	size of the collective remove requests for all the resources	*
*	of any given client will be a reasonable size.			*
*									*
************************************************************************/
#include <X11/X.h>
#include <X11/Xproto.h>

#include "xmx.h"
#include "resource.h"

static buffer_t buf;

/************************************************************************
*									*
*   xreq_flush								*
*									*
*	Broadcast all x requests collected so far.			*
*									*
************************************************************************/
void
xreq_flush()
{
   if (buf.n)
      broadcast(&buf);

   buf_clear(&buf, 0);	/* just to be sure... */
}

/************************************************************************
*									*
*   xreq_reparent_window						*
*									*
************************************************************************/
void
xreq_reparent_window(w, p, x, y)
   Window w, p;
   int x, y;
{
   static inited = 0;
   static xReparentWindowReq rep;

   if (!inited) {
      rep.reqType = X_ReparentWindow;
      rep.length = sz_xReparentWindowReq >> 2;
      inited = 1;
   }
   rep.window = w;
   rep.parent = p;
   rep.x = x;
   rep.y = y;

   buf_put(&buf, (char *)&rep, sz_xReparentWindowReq);
}

/************************************************************************
*									*
*   xreq_map_window							*
*									*
************************************************************************/
void
xreq_map_window(w)
   Window w;
{
   static inited = 0;
   static xResourceReq req;

   if (!inited) {
      req.reqType = X_MapWindow;
      req.length = sz_xResourceReq >> 2;
      inited = 1;
   }
   req.id = w;

   buf_put(&buf, (char *)&req, sz_xResourceReq);
}

/************************************************************************
*									*
*   xreq_destroy_window							*
*									*
************************************************************************/
void
xreq_destroy_window(w)
   Window w;
{
   static inited = 0;
   static xResourceReq req;

   if (!inited) {
      req.reqType = X_DestroyWindow;
      req.length = sz_xResourceReq >> 2;
      inited = 1;
   }
   req.id = w;

   buf_put(&buf, (char *)&req, sz_xResourceReq);
}

/************************************************************************
*									*
*   xreq_free_resource							*
*									*
************************************************************************/
void
xreq_free_resource(type, id)
   rtype_t type;
   rid_t id;
{
   static inited = 0;
   static xResourceReq req;

   switch (type) {
      case FONT:
         req.reqType = X_CloseFont;
         break;
      case PIXMAP:
         req.reqType = X_FreePixmap;
         break;
      case GC:
         req.reqType = X_FreeGC;
         break;
      case CMAP:
         req.reqType = X_FreeColormap;
         break;
      case CURSOR:
         req.reqType = X_FreeCursor;
         break;
   }

   if (!inited) {
      req.length = sz_xResourceReq >> 2;
      inited = 1;
   }
   req.id = id;

   buf_put(&buf, (char *)&req, sz_xResourceReq);
}

/************************************************************************
*									*
*   xreq_create_colormap						*
*									*
************************************************************************/
void
xreq_create_colormap(cmap, visual, window, alloc)
   rid_t cmap, visual, window;
   int alloc;
{
   static inited = 0;
   static xCreateColormapReq req;

   if (!inited) {
      req.reqType = X_CreateColormap;
      req.length = sz_xCreateColormapReq >> 2;
      inited = 1;
   }
   req.alloc = (BYTE)alloc;
   req.mid = (Colormap)cmap;
   req.window = (Window)window;
   req.visual = (VisualID)visual;

   buf_put(&buf, (char *)&req, sz_xCreateColormapReq);
}

/************************************************************************
*									*
*   xreq_change_window_colormap						*
*									*
************************************************************************/
void
xreq_change_window_colormap(window, cmap)
   rid_t window, cmap;
{
   static inited = 0;
   static xChangeWindowAttributesReq req;

   if (!inited) {
      req.reqType = X_ChangeWindowAttributes;
      req.length = (sz_xChangeWindowAttributesReq >> 2) + 1;
      req.valueMask = CWColormap;
      inited = 1;
   }
   req.window = window;

   buf_put(&buf, (char *)&req, sz_xChangeWindowAttributesReq);
   buf_put(&buf, (char *)&cmap, sizeof(rid_t));
}

/************************************************************************
*									*
*   xreq_install_colormap						*
*									*
************************************************************************/
void
xreq_install_colormap(cmap)
   rid_t cmap;
{
   static inited = 0;
   static xResourceReq req;

   if (!inited) {
      req.reqType = X_InstallColormap;
      req.length = sz_xResourceReq >> 2;
      inited = 1;
   }
   req.id = cmap;

   buf_put(&buf, (char *)&req, sz_xResourceReq);
}

/************************************************************************
*									*
*   xreq_uninstall_colormap						*
*									*
************************************************************************/
void
xreq_uninstall_colormap(cmap)
   rid_t cmap;
{
   static inited = 0;
   static xResourceReq req;

   if (!inited) {
      req.reqType = X_UninstallColormap;
      req.length = sz_xResourceReq >> 2;
      inited = 1;
   }
   req.id = cmap;

   buf_put(&buf, (char *)&req, sz_xResourceReq);
}

/************************************************************************
*									*
*   xreq_alloc_color_cells						*
*									*
************************************************************************/
void
xreq_alloc_color_cells(cmap, ncol)
   rid_t cmap;
   int ncol;
{
   static inited = 0;
   static xAllocColorCellsReq req;

   if (!inited) {
      req.reqType = X_AllocColorCells;
      req.length = sz_xAllocColorCellsReq >> 2;
      req.planes = 0;
      inited = 1;
   }
   req.contiguous = 0;
   req.cmap = (Colormap)cmap;
   req.colors = ncol;

   buf_put(&buf, (char *)&req, sz_xAllocColorCellsReq);
}

/************************************************************************
*									*
*   xreq_store_colors							*
*									*
************************************************************************/
void
xreq_store_colors(cmap, ncol, pixels, rgb, flags)
   rid_t cmap;
   int ncol;
   pixel_t *pixels;
   unsigned short *rgb;		/* 3 consecutive shorts per color */
   int flags;
{
   register i;
   static inited = 0;
   static xStoreColorsReq req;
   char fg = (char)flags;

   if (!inited) {
      req.reqType = X_StoreColors;
      inited = 1;
   }
   req.length = (sz_xStoreColorsReq >> 2) + 3 * ncol;
   req.cmap = (Colormap)cmap;

   buf_put(&buf, (char *)&req, sz_xStoreColorsReq);

   for (i=0; i<ncol; i++) {
      buf_put(&buf, (char *)pixels, sizeof(pixel_t));
      buf_put(&buf, (char *)rgb, 3 * sizeof(short));
      buf_put(&buf, (char *)&fg, 2 * sizeof(char));
      pixels++;
      rgb += 3;
   }
}

/************************************************************************
*									*
*   xreq_free_colors							*
*									*
************************************************************************/
void
xreq_free_colors(cmap, npixels, pixels)
   rid_t cmap;
   int npixels;
   pixel_t *pixels;
{
   static inited = 0;
   static xFreeColorsReq req;

   if (!inited) {
      req.reqType = X_FreeColors;
      req.planeMask = 0;
      inited = 1;
   }
   req.length = (sz_xFreeColorsReq >> 2) + npixels;
   req.cmap = cmap;

   buf_put(&buf, (char *)&req, sz_xFreeColorsReq);
   buf_put(&buf, (char *)pixels, npixels << 2);
}
