/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   main.c								*
*									*
*	Xmx is an X protocol multiplexor.  It acts as an intermediary	*
*	between clients and a controlling server and passes along	*
*	output requests to other servers, thus effectively		*
*	"broadcasting" the controlling server's graphics to n machines.	*
*									*
************************************************************************/
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <ctype.h>
#include <strings.h>
#include <stdio.h>
#include <errno.h>
#ifdef hpux
#include <unistd.h>
#define getdtablesize() sysconf(_SC_OPEN_MAX)
#endif

#include "xmx.h"

#define INITIAL_MESSAGE "XMX Version 1.0, ready."

#define USAGE	"usage: %s [:displaynum] [-option ...] [display ...]\n"
char *options[] = {
		"-display display\tmaster X server to connect to\n",
		"-f filename\t\tfile with list of slave X servers\n",
		"-i\t\t\tinteractive mode\n",
		"-S\t\t\tslave only mode\n",
		"-q\t\t\tquiet mode - don't print any messages\n",
		"-maxfd\t\t\tset NOFILE to maximum value\n",
		"-msb num\t\tmultiply slave sndbuf size by num\n",
		"-nocmap\t\t\tdon't allocate default colormap entries\n",
		"-help\t\t\tprint message with these options\n",
		(char *)0};

#define DISPLAY_EV	"DISPLAY"

connection_t	*cv;			/* connection info vector */
int		fdts;			/* file descriptor table size */
int		maxfd = -1;		/* largest fd in use */

int		debug = 0;
static histq_t	slavehist;		/* history of slave requests */

static void	new_conn();
static void	close_conn();
static int	split();
static void	usage();
static void	get_args();
static int	read_server_list();

main(argc, argv)
   int argc;
   char **argv;
{
   register i, r, s, t, rv;
   int interactive;
   int pointer;
   int hasmaster;
   int bfactor = 1;
   char *cp = 0, *dp;
   int *fdv;			/* array of file descriptors */
   int dpy;			/* display we listen on */
   int display, screen;
   int noslaves = 1;		/* were any slaves specified? */
   int masterfd;		/* socket of initial master connection */
   histq_t masterhist;		/* initial master connection recent requests */
   display_t master;		/* keep master connect info around */
   int grab = -1;		/* implements GrabServer request */
   char **slist;
   fd_set rfds, wfds, fdsave;
   buffer_t buf, *bp;
   struct timeval *tvp = 0;

#ifdef MALLOC_DEBUG
malloc_debug(2);
#endif
   get_args(argc,argv,&dpy,&interactive,&pointer,&hasmaster,&bfactor,&slist);

				/* allocate array of file descriptor info */
   fdts = getdtablesize();
   if ((cv = (connection_t *)calloc((unsigned)fdts, sizeof(connection_t))) == 0)
      quit(1, "calloc returned zero\n");

   buf_init(fdts);
   map_init(fdts);

   FD_ZERO(&fdsave);
				/* open client connections to servers */
   for (i=0; slist[i]; i++) {
      if (split(slist[i], &display, &screen))
         quit(1, "bad display string [%s]\n", slist[i]);
      if ((s = open_display(slist[i], display, (buffer_t *)0, bfactor)) < 0)
         if (i == 0 && hasmaster) {
            fprintf(stderr, "cannot connect to master server, exiting\n");
            exit(1);
         }
         else
            continue;

      if (i == 0 && hasmaster) {
         new_conn(s, MASTER, &fdsave);
         master.hostname = slist[i];
         master.display = display;
         master.screen = screen;
         masterfd = s;
         cv[s].u.master.mate = -1;
         if (pointer)
            pointer_init(screen, &tvp, 0);
      }
      else {
         new_conn(s, SLAVE, &fdsave);
         cv[s].u.slave.dpy.hostname = slist[i];
         cv[s].u.slave.dpy.display = display;
         cv[s].u.slave.dpy.screen = screen;
         if (noslaves) noslaves = 0;
      }
   }
   if (interactive) {
      new_conn(fileno(stdin), USER, &fdsave);
      new_conn(fileno(stdout), USER, &fdsave);
   }
   else
      fclose(stdin);
				/* listen on well-known ports like an Xserver*/
   if ((i = server_sockets(dpy, &fdv)) < 0)
      exit(1);
   for (t=0; t<i; t++)
      new_conn(fdv[t], CONNECT, &fdsave);

   request_hist_init(&masterhist);
   request_hist_init(&slavehist);
   if (!hasmaster) pointer = 0;
   buf_zero(&buf);

   printf("%s\n", INITIAL_MESSAGE);
   fflush(stdout);
   xreq_flush();		/* initial broadcast */

				/* main loop */
   while (1) {
      if (grab < 0) {			/* server not grabbed */
         bcopy((char *)&fdsave, (char *)&rfds, sizeof(fd_set));
         FD_ZERO(&wfds);
      }
      else {				/* only listen to grabbing client */
         FD_ZERO(&rfds);
         FD_SET(grab, &rfds);
         FD_SET(cv[grab].u.client.mate, &rfds);
      }
      buf_setwfds(maxfd, &wfds);
      r = select(maxfd+1, &rfds, &wfds, (fd_set *)0, tvp);
      if (r <= 0)
         if (r < 0)
            pquit(1, "select");
         else {
            if (pointer && grab<0)
               pointer_query(masterfd, &masterhist);
            continue;
         }
				/* flush out any writes we can now do */
      for (i=0; i<=maxfd; i++)
         if (FD_ISSET(i, &wfds))
            if (buf_flush(i, NBLOCK) < 0)
               perror("buf_flush");

      for (i=0; i<=maxfd; i++) {
         if (FD_ISSET(i, &rfds))
            switch (cv[i].type) {
               case USER:
#ifdef DEBUG
                  if (debug) fprintf(stderr,"==USER {%d}\n",i);fflush(stdout);
#endif
                  switch (getchar()) {
                     case 'q':
                        exit(0);
                     case 'd':
                        debug = debug ? 0 : 1;
                        break;
                  }
                  break;
               case CONNECT:
#ifdef DEBUG
                  if (debug)fprintf(stderr,"==CONNECT {%d}\n",i);fflush(stdout);
#endif
                  if ((s = accept_client(i)) < 0)	/* physically accept */
                     continue;
                  if (test_client_block(s))		/* read from client */
                     continue;
                  new_conn(s, CLIENT, &fdsave);
                  cv[s].u.client.state = resource_alloc_client(0);
                  request_hist_init(&cv[s].u.client.hist);
                  if (!hasmaster) {
                     bp = (buffer_t *)0;
                     cv[s].u.client.mate = -1;
                  }
                  else {				/* if master, open */
                     bp = &buf;
                     buf_clear(bp, 0);
                     t = open_display(master.hostname,master.display,bp,1);
                     if (t < 0) {
                        close_conn(s, &fdsave);
                        continue;
                     }
                     cv[s].u.client.mate = t;
                     new_conn(t, MASTER, &fdsave);
                     cv[t].u.master.mate = s;
                  }

                  if (send_setup(s, bp) < 0) {		/* respond to client */
                     close_conn(s, &fdsave);
                     close_conn(t, &fdsave);
                     continue;
                  }
                  break;
               case CLIENT:
#ifdef DEBUG
                  if (debug) fprintf(stderr,"==CLIENT {%d}\n",i);fflush(stdout);
#endif
                  bp = &cv[i].u.client.buf;
                  r = buf_read(i, bp, 0, NBLOCK);
/*warn("  CLIENT: read %d bytes\n", r);*/
                  if (bp->n) {
                     buf_clear(&buf, 0);
                     for (cp=(char *)0,t=0; s=request_get(bp, &cp); t+=s) {
                       request_hist_push(&cv[i].u.client.hist,request_type(cp));
                        if (hasmaster)
                           cv[cv[i].u.client.mate].u.master.seqnum++;
#ifdef DEBUG
                        if (debug)
                           request_print(cp, hasmaster?
                              cv[cv[i].u.client.mate].u.master.seqnum:0);
#endif
                        if (rv = request_grabserver(cp)) {
                           if (rv > 0) {
                              grab = i;	/* this could be more elegant... */
                              if (FD_ISSET(cv[grab].u.client.mate, &rfds)) {
                                 FD_ZERO(&rfds);
                                 FD_SET(cv[grab].u.client.mate, &rfds);
                              }
                              else
                                 FD_ZERO(&rfds);
                           }
                           else
                              grab = -1;
                        }
                        if (!noslaves) {
                           if (resource_request(cp,cv[i].u.client.state))
                              if (request_use(cp, 0))
                                 buf_put(&buf, cp, s);
                        }
                     }
                     if (!noslaves) {
                        xreq_flush();	/* kludge, kludge, kludge... */
                        broadcast(&buf);
                        map_rid_remove_marked();
                     }
                     if (t) {
                        if (!hasmaster)
                           buf_clear(bp, t);
                        else
                           if (buf_write(cv[i].u.client.mate, bp, t, BLOCK)<0) {
                              perror("master write");
                              continue;
                           }
                        if (pointer && grab<0)
                           pointer_query(masterfd, &masterhist);
                     }
                  }
                  if (r <= 0) {
                     if (r < 0)
                        perror("client read");
                     buf_clear(bp, 0);
                     resource_free(cv[i].u.client.state);
                     if (hasmaster) {
                        map_server_clear(cv[i].u.client.mate);
                        close_conn(cv[i].u.client.mate, &fdsave);
                     }
                     close_conn(i, &fdsave);
                     continue;
                  }
                  break;
               case MASTER:
#ifdef DEBUG
                  if (debug) fprintf(stderr,"==MASTER {%d}\n",i);fflush(stdout);
#endif
                  bp = &cv[i].u.master.buf;
                  r = buf_read(i, bp, 0, NBLOCK);
/*warn("  MASTER: read %d bytes\n", r);*/
                  if (bp->n) {
                     buf_clear(&buf, 0);
                     for (cp=(char *)0,t=0; s=reply_get(bp,&cp,(int*)0); t+=s) {
#ifdef DEBUG
                        if (debug)
                           if (i == masterfd)
                              reply_print(cp, &masterhist);
                           else
                              reply_print(cp,
                                      &cv[cv[i].u.master.mate].u.client.hist);
#endif
                        if (!noslaves)
                           if (i == masterfd) {
                              if (pointer)
                                 if (s = pointer_echo(cp, &masterhist, &dp))
                                    buf_put(&buf, dp, s);
                           }
                           else
                              resource_reply(cp,
				&cv[cv[i].u.master.mate].u.client.hist,
				cv[cv[i].u.master.mate].u.client.state);
                     }
                     if (!noslaves) {
                        xreq_flush();		/* REDUNDANT!!!! */
                        broadcast(&buf);
                     }
                     if (cv[i].u.master.mate < 0) {	/* no mate, no write */
                        buf_clear(bp, 0);
                        continue;
                     }
                     if (t)
                        if (buf_write(cv[i].u.master.mate, bp, t, BLOCK) < 0) {
                           perror("client write");
                           continue;
                        }
                  }
                  if (r <= 0) {
                     if (r < 0)
                        perror("master read");
                     buf_clear(bp, 0);
                     close_conn(i, &fdsave);
                     if (cv[i].u.master.mate >= 0) {
                        resource_free(cv[cv[i].u.master.mate].u.client.state);
                        close_conn(cv[i].u.master.mate, &fdsave);
                     }
                     continue;
                  }
                  break;
               case SLAVE:
#ifdef DEBUG
                  if (debug) fprintf(stderr,"==SLAVE {%d}\n",i);fflush(stdout);
#endif
                  buf_clear(&buf, 0);
#ifdef WHY_IS_THIS_HERE
                  r = buf_read(i, &buf, cv[i].u.slave.flush, NBLOCK);
#else
                  r = buf_read(i, &buf, 0, NBLOCK);
#endif
/*warn("  SLAVE: read %d bytes\n", r);*/
                  if (r <= 0) {
                     if (r < 0)
                        perror("slave read");
                     close_conn(i, &fdsave);
                     continue;
                  }
                  if (cv[i].u.slave.flush) { /* ignore and flush long replies */
                     cv[i].u.slave.flush -= r;
                     continue;
                  }
                  for (	cp=(char *)0,t=0;
			s=reply_get(&buf, &cp, &cv[i].u.slave.flush);
			t+=s)
#ifdef DEBUG
                     if (debug)
                        reply_print(cp, &slavehist);
#else
                     ;
#endif
                  break;
            }
      }
   }
}

/************************************************************************
*									*
*   new_conn								*
*									*
*	Set the type and the fd_set for the new connection.		*
*									*
************************************************************************/
static void
new_conn(s, type, fdsp)
   int s;
   short type;
   fd_set *fdsp;
{
   cv[s].type = type;
   FD_SET(s, fdsp);
   if (s > maxfd)
      maxfd = s;
}

/************************************************************************
*									*
*   close_conn								*
*									*
*	Close the file descriptor, clear the connection structure and	*
*	the fd_set.							*
*									*
************************************************************************/
static void
close_conn(s, fdsp)
   int s;
   fd_set *fdsp;
{
   close(s);
   cv[s].type = CLOSED;
   FD_CLR(s, fdsp);
   if (s == maxfd) {
      while (cv[--s].type == CLOSED);
      maxfd = s;
   }
}

/************************************************************************
*									*
*   broadcast								*
*									*
*	Write a buffer to all slave servers.				*
*									*
************************************************************************/
void
broadcast(bp)
   buffer_t *bp;
{
   register i, k, once = 1;
   char *cp, *dp;
   static sseq = 1;
   static buffer_t buf;

   buf_put(&buf, bp->cp, bp->n);	/* make a scratch copy */

   for (i=0; i<=maxfd; i++)
      if (cv[i].type == SLAVE) {
         cp = dp = (char *)0;
#ifdef DEBUG
         k = 0;
#endif
         while (request_get(bp, &dp) && request_get(&buf, &cp)) {
            (void)request_map(i, dp, cp);
#ifdef DEBUG
            if (once)
               request_hist_push(&slavehist, request_type(cp));
            if (debug) {
               fprintf(stderr,"TO SLAVE: ");
               request_print(cp, sseq + k++);
            }
#endif
         }
         if (buf_write(i, &buf, 0, NBLOCK) < 0) {
            perror("slave server write");
            fprintf(stderr, "closing connection\n");
            cv[i].type = CLOSED;
            close(i);
         }
         buf.n = bp->n;
         once = 0;
      }
#ifdef DEBUG
   sseq += k;
#endif

   buf_clear(&buf, 0);
}

/************************************************************************
*									*
*   split								*
*	Null terminate the hostname part of a display string		*
*	(in place) and pass back the display and screen numbers.	*
*									*
************************************************************************/
static int
split(str, d, s)
   char *str;
   int *d, *s;
{
   register char *cp, *dp;

   cp = index(str, ':');
   if (cp) {
      *cp++ = '\0';
      if (*cp) {
         if (*cp == ':') {
            fprintf(stderr, "DECnet not supported\n");
            return -1;
         }
         if (!isdigit(*cp))		/* missing display number */
            return -1;
         dp = index(cp, '.');
         if (dp) {
            if (!isdigit(*(dp+1)))	/* missing screen number */
               return -1;
            *dp++ = '\0';
            *s = atoi(dp);
         }
         else
            *s = 0;
         *d = atoi(cp);
      }
      else
         return -1;
   }
   else
      *d = *s = 0;

   return 0;
}

/************************************************************************
*									*
*   usage								*
*									*
*	Print usage message and options.				*
*									*
************************************************************************/
static void
usage(status, cmd)
   int status;
   char *cmd;
{
   register i;

   fprintf(stderr, USAGE, cmd);
   for (i=0; options[i]; i++)
      fprintf(stderr, options[i]);
   exit(status);
}

/************************************************************************
*									*
*   get_args								*
*									*
*	Process arglist.  Sets display number, flags and collects	*
*	list of servers to open.					*
*									*
************************************************************************/
static void
get_args(argc, argv, dpy, interactive, pointer, hasmaster, bfactor, slist)
   int argc;
   char **argv;
   int *dpy;
   int *interactive, *pointer, *hasmaster;
   int *bfactor;
   char ***slist;
   
{
   register i;
   register char *cp;
   register char *cmd = *argv;
   static char *cpv[1024];
   struct rlimit rlim;
				/* defaults */
   *dpy = -1;
   *interactive = 0;
   *pointer = 0;
   *hasmaster = 1;
   cpv[0] = cpv[1] = (char *)0;

   i = 1;
   while (--argc)
      if (**++argv  == '-')
         switch (*(*argv+1)) {
            case 'd':
               if (strcmp(*argv, "-display") == 0) {
                  argc--;
                  if (cpv[0])
                     usage(1, cmd);
                  cpv[0] = *++argv;
               }
               else if (strcmp(*argv, "-debug") == 0)
#ifdef DEBUG
                  debug = 1;
#else
                  fprintf(stderr, "Not compiled for debugging\n");
#endif
               else
                  usage(1, cmd);
               break;
            case 'f':
               argc--;
               i += read_server_list(*++argv, &cpv[i], 1024-i);
               break;
            case 'i':
               *interactive = 1;
               break;
            case 'p':
               *pointer = 1;
               break;
            case 'S':
               *hasmaster = 0;
               break;
            case 'q':
               errors(0);
               break;
            case 'm':
               if (strcmp(*argv, "-maxfd") == 0)
#ifdef RLIMIT_NOFILE
                  if (getrlimit(RLIMIT_NOFILE, &rlim) == 0) {
                     rlim.rlim_cur = rlim.rlim_max;
                     if (setrlimit(RLIMIT_NOFILE, &rlim))
                        warn("can't set NOFILE limit\n");
                  }
                  else
                     warn("can't query NOFILE limit\n");
#else
                  warn("file descriptor resource limit not settable\n");
#endif
               else if (strcmp(*argv, "-msb") == 0) {
                  if (--argc)
                     if (isdigit(**++argv)) {
                        *bfactor = atoi(*argv);
                        break;
                     }
                  usage(1, cmd);
                  break;
               }
               else
                  usage(1, cmd);
               break;
            case 'n':
               if (strcmp(*argv, "-nocmap") == 0)
                  resource_grab_cmap(0);
               else
                  usage(1, cmd);
               break;
            default:
               usage(strcmp(*argv, "-help") != 0, cmd);
         }
      else
         if (**argv == ':')		/* display we listen on */
            if (*dpy >= 0)
               usage(1, cmd);
            else
               *dpy = atoi(*argv+1);
         else				/* server to connect to */
            cpv[i++] = *argv;

				/* look for master in environment */
   if (*hasmaster && cpv[0] == 0)
      if (cp = (char *)getenv(DISPLAY_EV))
         cpv[0] = strcpy((char *)malloc((unsigned)(strlen(cp)+1)), cp);
      else
         quit(1, "no `master' display to open\n");

   cpv[i] = (char *)0;
   if (cpv[0])
      *slist = &cpv[0];
   else
      *slist = &cpv[1];
}

/************************************************************************
*									*
*   read_server_list							*
*									*
*	Read list of servers from file.  Malloc strings for each and	*
*	put them in the array given.					*
*									*
************************************************************************/
static int
read_server_list(file, cpv, n)
   char *file;
   char **cpv;
   int n;
{
   register i = 0;
   register int c;
   register char *cp;
   FILE *fp;
   char buf[1024];
   char *eobuf = buf+sizeof(buf);

   if (fp = fopen(file, "r"))
      for (cp=buf; fp;)
         switch (c = getc(fp)) {
            case EOF:
               fclose(fp);
               fp = (FILE *)0;
               /* fall through */
            case ' ':
            case '\t':
            case '\n':
               if (cp > buf) {
                  *cp = '\0';
                  if ((cpv[i] = (char *)malloc(strlen(buf) + 1)) == 0)
                     quit(1, "malloc returned zero\n");
                  (void)strcpy(cpv[i], buf);
                  if (++i >= n)
                     quit(1, "vector buffer overflow reading server list\n");
                  cp = buf;
               }
               break;
            default:
               *cp++ = (char)c;
               if (cp >= eobuf)
                  quit(1, "char buffer overflow reading server list\n");
               break;
         }
   else
      warn("can't open '%s'\n", file);

   cpv[i] = (char *)0;

   return i;
}
