/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   map.c								*
*									*
*	Provide a mapping between n client resource id spaces to	*
*	m server resource id spaces.  The client ids are assigned a	*
*	unique small integer and put into a hash table.  The integer	*
*	is used to construct a server resource id.			*
*	Also, map atoms.						*
*									*
************************************************************************/
#include <sys/param.h>

#include <X11/X.h>	/* needed for Xatom.h, next */
#include <X11/Xatom.h>	/* only need XA_LAST_PREDEFINED */

#include <stdio.h>

#include "xmx.h"

#define ATOMMAPSIZE	64	/* initial: atom maps grow as needed */
#define RIDMAPSIZE	1024

#define RHASH(rid,tsz)	(rid%tsz)	/* resource id hash function */

typedef struct _rmap_t {
	short		mark;		/* marked for removal */
	rid_t		rid;		/* client resource id */
	rid_t		srid;		/* assigned server rid index */
	char *		dptr;		/* data pointer */
	struct _rmap_t	*next;
}rmap_t;

typedef struct {
	int		shift;		/* # of low order zeros in mask */
	unsigned long	mask;		/* X server supplied mask */
	unsigned long	base;		/* X server supplied base */
	rmap_t		*ids;		/* X server initial resource ids */
	int		natoms;		/* number of slots in atom map */
	atom_t		*atoms;		/* atom map */
}server_t;

#ifndef RMAP_DEBUG
typedef struct me_t {
	rmap_t		*rmap;
	int		nuked;
	struct me_t	*next;
};
struct me_t	*rmap_debug_head = 0;
#endif

static int		master = -1;	/* master server */
static rmap_t *		rmap[RIDMAPSIZE];	/* must be zero initially */
static int		dirty;		/* does rmap have marked rids? */

static int		nservers;
static server_t **	servers;

/************************************************************************
*									*
*   map_init								*
*									*
*	Initialize map module -- allocate tables, etc.			*
*									*
************************************************************************/
void
map_init(fdts)
   int fdts;
{
   servers = (server_t **)calloc((unsigned)fdts, sizeof(server_t *));
   if (servers == 0)
      quit(1, "calloc returned zero\n");
   nservers = fdts;
}

/************************************************************************
*									*
*   map_server_init							*
*									*
*	Register an X server for resource id and atom mapping.		*
*									*
*	The first server registered with this routine is considered	*
*	the master server.  All server resource ids are mapped to and	*
*	from the master server's (i.e. clients see only master server	*
*	resource ids).							*
*									*
************************************************************************/
void
map_server_init(s, base, mask)
   register s;
   unsigned long base, mask;
{
   register n;

   if (s < 0 || s >= nservers)
      quit(1, "server index [%d] out of bounds\n", s);

   if (master < 0)
      master = s;
					/* allocate server block */
   if ((servers[s] = (server_t *)malloc(sizeof(server_t))) == 0)
      quit(1, "malloc returned zero\n");

					/* save resource id info */
   servers[s]->base = base;
   servers[s]->mask = mask;
   for (n=0; mask % 2 == 0; mask>>=1,n++);	/* # of rightmost zeros */
   servers[s]->shift = n;

   servers[s]->ids = (rmap_t *)0;
					/* allocate atom hash table */
   servers[s]->natoms = ATOMMAPSIZE;
   servers[s]->atoms = (atom_t *)calloc(ATOMMAPSIZE, sizeof(atom_t));
   if (servers[s]->atoms == 0)
      quit(1, "calloc returned zero\n");

}

/************************************************************************
*									*
*   map_server_rid_add							*
*									*
************************************************************************/
void
map_server_rid_add(s, rid, dptr)
   register s;
   rid_t rid;
   char *dptr;
{
   register rmap_t *rp;

   if ((rp = (rmap_t *)malloc(sizeof(rmap_t))) == 0)
      quit(1, "malloc returned zero\n");
#ifndef RMAP_DEBUG
{struct me_t *tmp = (struct me_t *)calloc(1,sizeof(struct me_t));
tmp->rmap = rp;
tmp->next = rmap_debug_head;
rmap_debug_head = tmp;
}
#endif

   rp->mark = 0;
   rp->rid = rid;
   rp->srid = NORID;
   rp->dptr = dptr;
   rp->next = servers[s]->ids;
   servers[s]->ids = rp;
}

/************************************************************************
*									*
*   map_server_clear							*
*									*
*	Clear all structures associated with the master server		*
*	connection given.						*
*									*
************************************************************************/
void
map_server_clear(s)
   register s;
{
   register rmap_t *lp, *rp;

   for (rp=servers[s]->ids; lp=rp;) {
      rp=rp->next;
      free((char *)lp);
#ifdef RMAP_DEBUG
{struct me_t tmp;
for (tmp=rmap_debug_head; tmp; tmp=tmp->next)
   if (tmp->rmap == lp) {
      if (tmp->nuked)
         warn("freeing a freed rmap!\n");
      else
         tmp->nuked = 1;
      break;
   }
}
#endif RMAP_DEBUG
   }
   if (servers[s]->atoms)
      free((char *)servers[s]->atoms);

   free((char *)servers[s]);

   servers[s] = (server_t *)0;
/*fprintf(stderr, "map_server_clear: all rids for %d cleared\n", s);*/
/*
fprintf(stderr, "---servers after map_server_clear---\n");
map_print_servers();
fprintf(stderr, "------------------------------------\n");
*/
}

/************************************************************************
*									*
*   map_rid_remove_marked						*
*									*
*	Remove all marked rids.						*
*									*
************************************************************************/
void
map_rid_remove_marked()
{
   register i;
   register rmap_t *lp, *rp, *tp;

   if (dirty) {
      for (i=0; i<RIDMAPSIZE; i++) {
         lp = (rmap_t *)0;
         for (rp=rmap[i]; tp=rp;) {
            rp=rp->next;

            if (tp->mark) {
               if (lp)
                  lp->next = tp->next;
               else
                  rmap[i] = tp->next;
               free((char *)tp);
#ifdef RMAP_DEBUG
{struct me_t tmp;
for (tmp=rmap_debug_head; tmp; tmp=tmp->next)
   if (tmp->rmap == tp) {
      if (tmp->nuked)
         warn("freeing a freed rmap!\n");
      else
         tmp->nuked = 1;
      break;
   }
}
#endif RMAP_DEBUG
            }
            else
               lp = tp;
         }
      }
      dirty = 0;
/*fprintf(stderr, "map_rid_remove_marked: done\n");*/
/*
fprintf(stderr, "---rmap after map_rid_remove_marked---\n");
map_print_rmap();
fprintf(stderr, "-----------------------------------------\n");
*/
   }
}

/************************************************************************
*									*
*   map_rid_mark							*
*									*
*	Mark a map table entry for removal later.			*
*									*
************************************************************************/
void
map_rid_mark(rid)
   rid_t rid;
{
   register i;
   register rmap_t *rp;

   i = RHASH(rid,nservers);

   for (rp=rmap[i]; rp && rid != rp->rid; rp=rp->next);

   if (rp) {
      rp->mark = 1;
      dirty = 1;
/*fprintf(stderr, "map_rid_mark: 0x%x\n", rid);*/
   }
   else
      warn("map_rid_mark: tried to mark nonexistent rid [0x%x]\n", rid);
}

/************************************************************************
*									*
*   map_rid_add								*
*									*
*	Insert a rid and associated data pointer into the hash table.	*
*									*
************************************************************************/
void
map_rid_add(rid, dptr)
   rid_t rid;
   char *dptr;
{
   register i;
   register rmap_t *rp;
   static srid = 0;

   i = RHASH(rid,nservers);

   for (rp=rmap[i]; rp && rid != rp->rid; rp=rp->next);

   if (rp)
      warn("map_rid_add: can't add duplicate rid [0x%x]\n", rid);
   else {
      if ((rp = (rmap_t *)malloc(sizeof(rmap_t))) == 0)
         quit(1, "malloc returned zero\n");
#ifndef RMAP_DEBUG
{struct me_t *tmp = (struct me_t *)calloc(1,sizeof(struct me_t));
tmp->rmap = rp;
tmp->next = rmap_debug_head;
rmap_debug_head = tmp;
}
#endif
      rp->mark = 0;
      rp->rid = rid;
      rp->srid = ++srid;
      rp->dptr = dptr;
      rp->next = rmap[i];
      rmap[i] = rp;
/*fprintf(stderr, "map_rid_add: 0x%x 0x%x\n", rid, dptr);*/
   }
}

/************************************************************************
*									*
*   map_rid								*
*									*
*	Map a real client resource id to a slave server rid, properly	*
*	shifted and or'd for the destination X server's consumption.	*
*									*
************************************************************************/
rid_t
map_rid(rid, s)
   rid_t rid;
   register s;		/* target server */
{
   register i;
   register rmap_t *rp, *sp;

   sp = servers[s]->ids;
   for (rp=servers[master]->ids; rp && rid != rp->rid; rp=rp->next)
      if (sp) sp = sp->next;
   if (rp && sp)
      return sp->rid;

   i = RHASH(rid,nservers);

   for (rp=rmap[i]; rp && rid != rp->rid; rp=rp->next);
/*
if (rp)
fprintf(stderr, "map_rid: found 0x%x %d\n", rid, rp->srid);
else
*/
if (!rp)
fprintf(stderr, "map_rid: not found 0x%x\n", rid);

   if (rp)
      return ((rp->srid<<servers[s]->shift)&servers[s]->mask)|servers[s]->base;
   else
      return NORID;
}

/************************************************************************
*									*
*   map_rid_data							*
*									*
*	Get data associated with a real client resource id.		*
*									*
************************************************************************/
char *
map_rid_data(rid)
   rid_t rid;
{
   register i;
   register rmap_t *rp;

   for (rp=servers[master]->ids; rp && rid != rp->rid; rp=rp->next);
   if (rp)
      return rp->dptr;

   i = RHASH(rid,nservers);

   for (rp=rmap[i]; rp && rid != rp->rid; rp=rp->next);

   if (rp)
      return rp->dptr;
   else
      return 0;
}

/************************************************************************
*									*
*   map_rid_unmap							*
*									*
*	A VERY inefficient routine, returns the original client rid	*
*	from a slave server rid.  Does a simple linear search of the	*
*	hash table.  For debugging purposes only.			*
*									*
************************************************************************/
rid_t
map_rid_unmap(rid, s)
   rid_t rid;
   register s;
{
   register i;
   register rid_t srid;
   register rmap_t *rp, *sp;

   rp = servers[master]->ids;
   for (sp=servers[s]->ids; sp && rid != sp->rid; sp=sp->next)
      if (rp) rp = rp->next;
   if (sp && rp)
      return rp->rid;

   srid = (unsigned int)((rid & ~servers[s]->base) >> servers[s]->shift);

   for (i=0; i<RIDMAPSIZE; i++) {
      for (rp=rmap[i]; rp && srid != rp->srid; rp=rp->next);
      if (rp)
         return rp->rid;
   }

   return NORID;
}

/************************************************************************
*									*
*   map_atom_add							*
*									*
*	Setup an atom mapping for the given slave server.		*
*									*
************************************************************************/
void
map_atom_add(atom, satom, s)
   atom_t atom, satom;
   register s;
{
   register atom_t a;
   register server_t *sp;

   if (atom <= XA_LAST_PREDEFINED)
      return;

   if (sp = servers[s]) {
      a = atom - (XA_LAST_PREDEFINED + 1);
      while (a >= sp->natoms) {
         sp->natoms += ATOMMAPSIZE;
         sp->atoms = (atom_t *)realloc(sp->atoms, sp->natoms*sizeof(atom_t));
         if (sp->atoms == 0)
            quit(1, "map_atom_add: atom [%d] too large to map\n", atom);
      }
      sp->atoms[a] = satom;
   }
}

/************************************************************************
*									*
*   map_atom								*
*									*
*	Return a slave server atom given a master server atom.		*
*									*
************************************************************************/
atom_t
map_atom(atom, s)
   atom_t atom;
   register s;
{
   register atom_t a;

   if (servers[s]) {
      a = atom - (XA_LAST_PREDEFINED + 1);
      return servers[s]->atoms[a];
   }

   return NOATOM;
}

/************************************************************************
*									*
*   map_print_servers							*
*									*
************************************************************************/
void
map_print_servers()
{
   register i;
   register rmap_t *rp;

   for (i=0; i<nservers; i++)
      if (servers[i]) {
          fprintf(stderr, "%02d 0x%x shift(%d) mask(0x%x) base(0x%x)",
                        i, servers[i], servers[i]->shift,
                        servers[i]->mask, servers[i]->base);
          if (i == master) fprintf(stderr, " MASTER");
          fprintf(stderr, "\n");
          fprintf(stderr, "\tids( ");
          for (rp=servers[i]->ids; rp; rp=rp->next)
             fprintf(stderr, "0x%x ", rp->rid);
          fprintf(stderr, ")\n");
      }
}

/************************************************************************
*									*
*   map_print_rmap							*
*									*
************************************************************************/
void
map_print_rmap()
{
   register i;
   register rmap_t *rp;

   for (i=0; i<RIDMAPSIZE; i++)
      for (rp=rmap[i]; rp; rp=rp->next)
         fprintf(stderr, "%02d rid(0x%x) srid(0x%x) dptr(0x%x)\n",
			i, rp->rid, rp->srid, rp->dptr);
}
