/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   resource.c								*
*									*
*	Routines to maintain a history of resource allocation per	*
*	client for use at connection close time.			*
*									*
************************************************************************/
#include <X11/X.h>
#define NEED_REPLIES
#include <X11/Xproto.h>

#include "xmx.h"
#include "resource.h"

static window_t *root;
static rid_t	last_cmap;
static int	grab_cmap = 1;

static void	set_events();
static void	remove_client_windows();
static void	reparent_savesets();
static void	link_window();
static void	unlink_window();
static void	free_subwindows();
static void	remove_resource();
static void	print_subtree();

void
resource_grab_cmap(bool)
   int bool;
{
   grab_cmap = bool;
}

/************************************************************************
*									*
*   resource_alloc_client						*
*									*
*	Allocate a new clientstate block.				*
*									*
************************************************************************/
clientstate_t *
resource_alloc_client(screen)
   int screen;
{
   rid_t rid;
   window_t *wp;
   register clientstate_t *c;

   c = (clientstate_t *)malloc(sizeof(clientstate_t));
   rid = rootwindow(screen);
   for (wp=root; wp && wp->rid != rid; wp=wp->nextsib);
   if (wp)
      c->root = wp;
   else {
      c->root = root;
      warn("resource_alloc_client: can't find root win for screen %d\n",screen);
   }
   c->closedownmode = DestroyAll;
   c->resources = (resource_t *)0;

   return c;
}

/************************************************************************
*									*
*   resource_new_root							*
*									*
*	Add a root window to the resource hierarchy model.		*
*									*
************************************************************************/
void
resource_new_root(s, rid, visualid, cmap, ncols, white, black)
   register s;
   rid_t rid, visualid, cmap;
   int ncols;
   pixel_t white, black;
{
   window_t *wp;
   pixel_t p[2];
   unsigned short rgb[6];

   if (root) {
      for (wp=root; wp->nextsib; wp=wp->nextsib);
      wp = wp->nextsib = (window_t *)calloc(1, sizeof(window_t));
   }
   else
      wp = root = (window_t *)calloc(1, sizeof(window_t));

   if (wp == 0)
      quit(1, "calloc returned zero\n");

   wp->rid = rid;
   wp->x = 0;
   wp->y = 0;
   wp->flags |= MAPPED;

   map_server_rid_add(s, rid, (char *)wp);
   map_server_rid_add(s, cmap, (char *)0);
   if (grab_cmap)
      xreq_alloc_color_cells(cmap, ncols-2);	/* hack */
}

/************************************************************************
*									*
*   resource_request							*
*									*
*	Interpret a protocol request in terms of resources allocated.	*
*									*
*	It's possible that a client could send a request that is	*
*	acceptable to this routine, but that causes a protocol error.	*
*	If the client continues processing, the resource history	*
*	maintained below will be inconsistent with the servers'.	*
*	I suppose we could keep an eye on server errors and implement	*
*	an undo...							*
*									*
************************************************************************/
int
resource_request(cp, client)
   char *cp;
   clientstate_t *client;
{
   register i;
   register long *lp;
   register window_t *wp, *parent;
   register saveset_t *sp, *lsp;
   register mask_t tm;

   switch (((xReq *)cp)->reqType) {
      case X_CreateWindow:
         {  register xCreateWindowReq *req = (xCreateWindowReq *)cp;

            parent=(window_t *)map_rid_data(req->parent);
            if (parent == 0)
               return 0;
            if ((wp = (window_t *)calloc(1, sizeof(window_t))) == 0)
               quit(1, "malloc returned zero\n");
            wp->rid = req->wid;
            wp->x = parent->x + req->x;
            wp->y = parent->y + req->y;
            set_events(client, wp, req->mask, cp+sz_xCreateWindowReq);
            link_window(parent, wp);
            map_rid_add(wp->rid, (char *)wp);
            wp->client = client;
         }
         break;
      case X_ChangeWindowAttributes:
         {  register xChangeWindowAttributesReq *req =
					(xChangeWindowAttributesReq *)cp;

            if (wp=(window_t *)map_rid_data(req->window))
               set_events(client, wp, req->valueMask,
					cp+sz_xChangeWindowAttributesReq);
         }
         break;
      case X_DestroyWindow:
         {  register xResourceReq *req = (xResourceReq *)cp;

            wp=(window_t *)map_rid_data(req->id);
            if (wp == 0)
               return 0;
            if (wp->parent == 0)
               return 0;    /* tried to destroy root window (not an error) */
            unlink_window(wp);
            wp->nextsib = (window_t *)0;
            free_subwindows(wp);
         }
         break;
      case X_DestroySubwindows:
         {  register xResourceReq *req = (xResourceReq *)cp;

            wp=(window_t *)map_rid_data(req->id);
            if (wp == 0)
               return 0;
            if (wp->child) {
               free_subwindows(wp->child);
               wp->child = (window_t *)0;
            }
         }
         break;
      case X_ChangeSaveSet:
         {  register xChangeSaveSetReq *req = (xChangeSaveSetReq *)cp;

            wp=(window_t *)map_rid_data(req->window);
            if (wp == 0)
               return 0;
            if (req->mode == 0) {	/* Insert */
               for (sp=wp->saveset; sp && sp->client!=client; sp=sp->next);
               if (sp == 0) {
                  if ((sp = (saveset_t *)malloc(sizeof(saveset_t))) == 0)
                     quit(1, "malloc returned zero\n");
                  sp->client = client;
                  sp->next = wp->saveset;
                  wp->saveset = sp;
               }
            }
            else {						/* Delete */
               lsp = (saveset_t *)0;
               for (sp=wp->saveset; sp && sp->client!=client; sp=sp->next)
                  lsp = sp;
               if (sp == 0) return 0;		/* not found */
               if (lsp == 0)
                  wp->saveset = sp->next;
               else
                  lsp->next = sp->next;
               free((char *)sp);
            }
         }
         break;
      case X_ReparentWindow:
         {  register xReparentWindowReq *req = (xReparentWindowReq *)cp;

            wp=(window_t *)map_rid_data(req->window);
            if (wp == 0)
               return 0;
            if (wp->parent == 0)
               return err(0, "tried to reparent root window [%x]\n", wp->rid);
            if ((parent=(window_t *)map_rid_data(req->parent)) == 0)
               return 0;
            wp->x = parent->x + req->x;
            wp->y = parent->y + req->y;
            unlink_window(wp);
            link_window(parent, wp);
         }
         break;
      case X_MapWindow:
         {  register xResourceReq *req = (xResourceReq *)cp;

            if (wp=(window_t *)map_rid_data(req->id)) {
               if (!(wp->flags & OVERRIDE_REDIRECT))
                  if (parent = wp->parent)
                     if (	(parent->flags & SUBSTRUCTURE_REDIRECT) &&
				parent->substr_c != client		)
                        return 0;
            }
            else
               return 0;
            wp->flags |= MAPPED;
         }
         break;
      case X_MapSubwindows:
         {  register xResourceReq *req = (xResourceReq *)cp;

            if (wp=(window_t *)map_rid_data(req->id)) {
			/****************************************
			* This is wrong.  Really we have to	*
			* break up this request into individual	*
			* MapWindow requests and check the	*
			* masks for each window.		*
			****************************************/
               if (	(wp->flags & SUBSTRUCTURE_REDIRECT) &&
			wp->substr_c != client			)
                  return 0;
            }
            else
               return 0;
         }
         break;
      case X_UnmapWindow:
         {  register xResourceReq *req = (xResourceReq *)cp;

            wp=(window_t *)map_rid_data(req->id);
            if (wp == 0)
               return 0;
            if (wp->parent == 0)	/* can't unmap root window */
               return 0;
            wp->flags &= ~MAPPED;
         }
         break;
      case X_UnmapSubwindows:
         {  register xResourceReq *req = (xResourceReq *)cp;

            wp=(window_t *)map_rid_data(req->id);
            if (wp == 0)
               return 0;
         }
         break;
      case X_ConfigureWindow:
         {  register xConfigureWindowReq *req = (xConfigureWindowReq *)cp;

            if (wp=(window_t *)map_rid_data(req->window)) {

				/*  catch changes in window location */
               if (wp->parent) {
                  lp = (long *)(req + 1);
                  for (tm=req->mask, i=0; tm; tm>>=1, i++)
                     if ((1<<i) & CWX & req->mask)
                        wp->x = wp->parent->x + (short)lp[i];
                     else if ((1<<i) & CWY & req->mask)
                        wp->y = wp->parent->y + (short)lp[i];
               }

				/* don't send redirected requests to slaves */
               if (!(wp->flags & OVERRIDE_REDIRECT))
                  if (parent = wp->parent)
                     if (	(parent->flags & SUBSTRUCTURE_REDIRECT) &&
				parent->substr_c != client		)
                        return 0;
               if (	(wp->flags & RESIZE_REDIRECT) &&
			(req->mask & (CWWidth | CWHeight)) &&
			wp->resize_c != client			)
                  return 0;
            }
            else
               return 0;
         }
         break;
      case X_OpenFont:
         {  register xOpenFontReq *req = (xOpenFontReq *)cp;

            resource_add(req->fid, FONT, client);
            map_rid_add(req->fid, (char *)0);
         }
         break;
      case X_CloseFont:
         {  register xResourceReq *req = (xResourceReq *)cp;

            remove_resource(req->id, FONT, client);
         }
         break;
      case X_CreatePixmap:
         {  register xCreatePixmapReq *req = (xCreatePixmapReq *)cp;

            resource_add(req->pid, PIXMAP, client);
            map_rid_add(req->pid, (char *)0);
         }
         break;
      case X_FreePixmap:
         {  register xResourceReq *req = (xResourceReq *)cp;

            remove_resource(req->id, PIXMAP, client);
         }
         break;
      case X_CreateGC:
         {  register xCreateGCReq *req = (xCreateGCReq *)cp;

            resource_add(req->gc, GC, client);
            map_rid_add(req->gc, (char *)0);
         }
         break;
      case X_FreeGC:
         {  register xResourceReq *req = (xResourceReq *)cp;

            remove_resource(req->id, GC, client);
         }
         break;
      case X_CreateColormap:
         {  register xCreateColormapReq *req = (xCreateColormapReq *)cp;

            resource_add(req->mid, CMAP, client);
            map_rid_add(req->mid, (char *)0);
#ifdef OUT_OF_SEQUENCE_BUG
            xreq_create_colormap(req->mid, req->visual, req->window, AllocAll);
#endif
         }
         break;
      case X_FreeColormap:
         {  register xResourceReq *req = (xResourceReq *)cp;

            remove_resource(req->id, CMAP, client);
         }
         break;
      case X_CopyColormapAndFree:
         {  register xCopyColormapAndFreeReq *req=(xCopyColormapAndFreeReq *)cp;

            resource_add(req->mid, CMAP, client);
            map_rid_add(req->mid, (char *)0);
            /** create a new one, query the old one, store all colors in new **/
         }
         break;
      case X_AllocColor:
         {  register xAllocColorReq *req = (xAllocColorReq *)cp;

            last_cmap = req->cmap;
         }
         return 0;
      case X_AllocNamedColor:
         {  register xAllocNamedColorReq *req = (xAllocNamedColorReq *)cp;

            last_cmap = req->cmap;
         }
         return 0;
      case X_AllocColorCells:
         return 0;
      case X_AllocColorPlanes:
         return 0;
      case X_FreeColors:
         return 0;
      case X_CreateCursor:
         {  register xCreateCursorReq *req = (xCreateCursorReq *)cp;

            resource_add(req->cid, CURSOR, client);
            map_rid_add(req->cid, (char *)0);
         }
         break;
      case X_CreateGlyphCursor:
         {  register xCreateGlyphCursorReq *req = (xCreateGlyphCursorReq *)cp;

            resource_add(req->cid, CURSOR, client);
            map_rid_add(req->cid, (char *)0);
         }
         break;
      case X_FreeCursor:
         {  register xResourceReq *req = (xResourceReq *)cp;

            remove_resource(req->id, CURSOR, client);
         }
         break;
      case X_QueryExtension:
         {  register xQueryExtensionReq *req = (xQueryExtensionReq *)cp;
         }
         break;
      case X_ListExtensions:
         {  register xReq *req = (xReq *)cp;
         }
         break;
      case X_SetCloseDownMode:
         break;
      case X_KillClient:
         break;
   }
   return 1;
}

void
resource_reply(cp, hp, client)
   char *cp;
   histq_t *hp;
   clientstate_t *client;
{
   register xGenericReply *reply = (xGenericReply *)cp;

   switch (reply->type) {
      case X_Error:
         return;
      case X_Reply:
         switch (request_hist_get(hp, reply->sequenceNumber)) {
            case X_AllocColor:
               xreq_store_colors(last_cmap, 1,
				&((xAllocColorReply *)cp)->pixel,
				&((xAllocColorReply *)cp)->red,
				DoRed | DoGreen | DoBlue);
               break;
            case X_AllocNamedColor:
               xreq_store_colors(last_cmap, 1,
				&((xAllocNamedColorReply *)cp)->pixel,
				&((xAllocNamedColorReply *)cp)->screenRed,
				DoRed | DoGreen | DoBlue);
               break;
            case X_QueryExtension:		/* a temporary hack */
               ((xQueryExtensionReply *)cp)->present = xFalse;
               break;
         }
         break;
      default:
         break;
   }
}
/************************************************************************
*									*
*   set_events								*
*									*
*	Scan attribute and event masks and set flag bits.		*
*									*
************************************************************************/
static void
set_events(client, wp, mask, list)
   clientstate_t *client;	/* requesting client */
   window_t *wp;		/* window affected */
   mask_t mask;			/* request value mask */
   long *list;			/* request value list */
{
   register mask_t tm;
   register mask_t maxval = CWOverrideRedirect | CWEventMask;

   for (tm=1; tm<=maxval; tm<<=1)
      if (tm & mask) {
         if (tm == CWOverrideRedirect)
            if (*list)
               wp->flags |= OVERRIDE_REDIRECT;
            else
               wp->flags &= ~OVERRIDE_REDIRECT;
         else if (tm == CWEventMask) {
            if (*list & ResizeRedirectMask) {
               wp->flags |= RESIZE_REDIRECT;
               wp->resize_c = client;
            }
            else if (wp->resize_c == client) {
               wp->flags &= ~RESIZE_REDIRECT;
               wp->resize_c = (clientstate_t *)0;
            }

            if (*list & SubstructureRedirectMask) {
               wp->flags |= SUBSTRUCTURE_REDIRECT;
               wp->substr_c = client;
            }
            else if (wp->substr_c == client) {
               wp->flags &= ~SUBSTRUCTURE_REDIRECT;
               wp->substr_c = (clientstate_t *)0;
            }
         }
         list++;
      }
}

/************************************************************************
*									*
*   resource_free							*
*									*
*	Generate resource free requests for all resources that should	*
*	be freed when the client specified by "client" is killed.	*
*									*
************************************************************************/
void
resource_free(client)
   clientstate_t *client;
{
   register window_t *wp;
   register resource_t *rp, *trp;

   if (client->closedownmode == DestroyAll) {
      for (wp=client->root; wp; wp=wp->nextsib)
         if (wp->child) {
            reparent_savesets(wp->child, client);
            remove_client_windows(wp->child, client);
         }

      for (rp=client->resources; trp=rp;) {
         map_rid_mark(rp->rid);
         xreq_free_resource(rp->type, rp->rid);
         rp=rp->next;
         free((char *)trp);
      }
      free((char *)client);
   }
   xreq_flush();
   map_rid_remove_marked();
}

/************************************************************************
*									*
*   remove_client_windows						*
*									*
************************************************************************/
static void
remove_client_windows(wp, client)
   register window_t *wp;
   clientstate_t *client;
{
   if (wp->nextsib)
      remove_client_windows(wp->nextsib, client);

   if (wp->client == client) {
      xreq_destroy_window(wp->rid);
      unlink_window(wp);
      wp->nextsib = (window_t *)0;
      free_subwindows(wp);
   }
   else if (wp->child)
      remove_client_windows(wp->child, client);
}

/************************************************************************
*									*
*   reparent_savesets							*
*									*
*	Recursively traverse the window hierarchy starting from the	*
*	given window and look for windows in our save set.  Reparent	*
*	them below their nearest ancestors that are not inferiors of	*
*	of any of our windows.						*
*									*
************************************************************************/
static void
reparent_savesets(wp, client)
   register window_t *wp;
   clientstate_t *client;
{
   register saveset_t *sp;
   register window_t *parent, *oa;	/* oa is oldest ancestor */

   if (sp = wp->saveset) {	/* if this window is part of a saveset */
      for (; sp && sp->client!=client; sp=sp->next);
      if (sp) {			/* ...and part of our saveset */
         oa = (window_t *)0;
         for (parent=wp->parent; parent; parent=parent->parent)
            if (parent->client == client)
               oa = parent;
         if (oa != 0) {		/* in save set AND an inferior of this client */
            xreq_reparent_window(wp->rid,
				oa->parent->rid,
				wp->x - oa->parent->x,
				wp->y - oa->parent->y);
            if (!(wp->flags & MAPPED))
               xreq_map_window(wp->rid);
            unlink_window(wp);		/* reparent it */
            link_window(oa->parent, wp);
            return;
         }
      }
   }
   if (wp->child)
      reparent_savesets(wp->child, client);

   if (wp->nextsib)
      reparent_savesets(wp->nextsib, client);
}

/************************************************************************
*									*
*   link_window								*
*									*
************************************************************************/
static void
link_window(parent, wp)
   register window_t *parent, *wp;
{
   wp->prevsib = (window_t *)0;
   wp->nextsib = parent->child;
   if (parent->child)
      parent->child->prevsib = wp;
   parent->child = wp;
   wp->parent = parent;
}

/************************************************************************
*									*
*   unlink_window							*
*									*
************************************************************************/
static void
unlink_window(wp)
   register window_t *wp;
{
   register window_t *parent;

   parent = wp->parent;
   if (parent->child == wp)
      if (wp->nextsib) {
         parent->child = wp->nextsib;
         wp->nextsib->prevsib = (window_t *)0;
      }
      else
         parent->child = (window_t *)0;
   else {
      if (wp->prevsib)
         wp->prevsib->nextsib = wp->nextsib;
      if (wp->nextsib)
         wp->nextsib->prevsib = wp->prevsib;
   }
}

/************************************************************************
*									*
*   free_subwindows							*
*									*
*	Recursively remove this and all child and nextsib windows.	*
*									*
************************************************************************/
static void
free_subwindows(wp)
   window_t *wp;
{
   register saveset_t *s, *t;

   if (wp->child)
      free_subwindows(wp->child);

   if (wp->nextsib)
      free_subwindows(wp->nextsib);

   map_rid_mark(wp->rid);
   for (s=wp->saveset; t=s;) {
      s = s->next;
      free((char *)t);
   }
   free((char *)wp);
}

/************************************************************************
*									*
*   resource_add							*
*									*
************************************************************************/
void
resource_add(rid, type, client)
   rid_t rid;
   rtype_t type;
   clientstate_t *client;
{
   resource_t *rp;

   if (client->resources)
      for (rp=client->resources; rp; rp=rp->next)
         if (rp->type == type && rp->rid == rid)
            return;
   if ((rp = (resource_t *)malloc(sizeof(resource_t))) == 0)
      quit(1, "malloc returned zero\n");
   rp->type = type;
   rp->rid = rid;
   rp->next = client->resources;
   client->resources = rp;
}

/************************************************************************
*									*
*   remove_resource							*
*									*
************************************************************************/
static void
remove_resource(rid, type, client)
   rid_t rid;
   rtype_t type;
   clientstate_t *client;
{
   resource_t *rp, *lrp;

   lrp = (resource_t *)0;
   for (rp=client->resources; rp; rp=rp->next) {
      if (rp->type == type && rp->rid == rid)
         break;
      lrp = rp;
   }
   if (rp == 0) return;
   if (lrp == 0)
      client->resources = rp->next;
   else
      lrp->next = rp->next;

   map_rid_mark(rid);
   free((char *)rp);
}

/************************************************************************
*									*
*   resource_print							*
*									*
*	Print resource heirarchy to stdout.  Debug routine.		*
*									*
************************************************************************/
void
resource_print()
{
   print_subtree(root, 0);
}

static void
print_subtree(wp, level)
   window_t *wp;
   int level;
{
   register i;
   saveset_t *sp;

   for (i=0; i<level; i++) printf("   ");

   printf("0x%x rid(0x%x) %c%c%c%c client(0x%x) saveset(", wp, wp->rid,
		(wp->flags & MAPPED) ? 'M' : '-',
		(wp->flags & SUBSTRUCTURE_REDIRECT) ? 'S' : '-',
		(wp->flags & OVERRIDE_REDIRECT) ? 'O' : '-',
		(wp->flags & RESIZE_REDIRECT) ? 'R' : '-',
		wp->client);

   for (sp=wp->saveset; sp; sp=sp->next)
      printf(" 0x%x", sp->client);

   printf(" )\n");

   if (wp->child)
      print_subtree(wp->child, level+1);
   if (wp->nextsib)
      print_subtree(wp->nextsib, level);
}
