/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   socket.c								*
*									*
*	All socket setup and io system calls are done here.		*
*									*
************************************************************************/
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#ifdef SYSV
#include <sys/file.h>
#endif

#include "xmx.h"

#define X_UNIX_DIR	"/tmp/.X11-unix"
#define X_UNIX_PATH	"/tmp/.X11-unix/X"
#define X_TCP_PORT	6000

/************************************************************************
*									*
*   server_sockets							*
*									*
*	Create X server style sockets, bind to well known addresses	*
*	and listen on them.  Returns the number of sockets or -1 if	*
*	an error occurs.   Fdsp is pointed to a static array of the	*
*	descriptors.							*
*									*
************************************************************************/
int
server_sockets(display, fdsp)
   int display;
   int **fdsp;
{
   register i;
   int s;
   struct linger linger;
   static int fds[2];			/* return array */
   static struct sockaddr_un usock;
   static struct sockaddr_in isock;

					/* create UNIX domain socket */
   usock.sun_family = AF_UNIX;
   (void)sprintf(usock.sun_path, "%s%d", X_UNIX_PATH, display);
   (void)mkdir(X_UNIX_DIR, 0777);
   (void)unlink(usock.sun_path);
   if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
      return perr(-1, usock.sun_path);

   i = strlen(usock.sun_path) + sizeof(usock.sun_family);
   if (bind(s, (struct sockaddr *)&usock, i))
      return perr(-1, usock.sun_path);

   if (listen(s, 5))
      return perr(-1, usock.sun_path);

   fds[0] = s;
					/* create TCP socket */
   isock.sin_family = AF_INET;
   isock.sin_port = htons((u_short)(X_TCP_PORT + display));
   isock.sin_addr.s_addr = htonl(INADDR_ANY);
   if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      return perr(-1, "tcp socket");

   for (i=20; i && bind(s, (struct sockaddr *)&isock, sizeof(isock)); i--)
      sleep(10);
   if (i == 0)
      return perr(-1, "tcp socket");

   linger.l_onoff = 0;
   linger.l_linger = 0;
   if (setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&linger, sizeof(linger)))
      return perr(-1, "tcp socket");

   if (listen(s, 5))
      return perr(-1, "tcp socket");

   fds[1] = s;

   *fdsp = fds;
   return 2;
}

/************************************************************************
*									*
*   accept_client							*
*									*
*	Accept a pending connection from a client.  Returns the new	*
*	socket file descriptor or -1 if an error occurred.		*
*									*
************************************************************************/
int
accept_client(s)
   int s;
{
   int i;
   int ns;
   int addrlen;
   struct sockaddr addr;

   addrlen = sizeof(addr);
   if ((ns = accept(s, &addr, &addrlen)) < 0)
      return perr(-1, "accept");

   if (addr.sa_family == AF_INET) {
      i = 1;
      if (setsockopt(ns, IPPROTO_TCP, TCP_NODELAY, (char *)&i, sizeof(i)))
         return perr(-1, "client socket");
   }
#ifdef FIOSNBIO
   {
      int arg = 1;
      ioctl(ns, FIOSNBIO, &arg);
   }
#else
   (void)fcntl(ns, F_SETFL, FNDELAY);
#endif

   return ns;
}

/************************************************************************
*									*
*   client_socket							*
*									*
*	Connect to an X server.  Returns the socket file descriptor or	*
*	-1 if an error occurred.					*
*									*
************************************************************************/
int
client_socket(hostname, display, bfactor)
   char *hostname;
   int display;
   int bfactor;
{
   register len;
   int s;
   int sendbufopt, sendbuflen;
   unsigned long iaddr;
   struct hostent *h;
   struct sockaddr *sock;
   static struct sockaddr_un usock;
   static struct sockaddr_in isock;

						/* UNIX domain socket */
   if (*hostname == '\0' || strcmp(hostname, "unix") == 0) {
      usock.sun_family = AF_UNIX;
      sprintf(usock.sun_path, "%s%d", X_UNIX_PATH, display);
      if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
         return perr(-1, usock.sun_path);

      sock = (struct sockaddr *)&usock;
      len = sizeof(usock);
   }
   else {					/* TCP socket */
      isock.sin_family = AF_INET;
      isock.sin_port = htons((u_short)(X_TCP_PORT + display));
      if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
         return perr(-1, "tcp socket");

      if ((long)(iaddr = inet_addr(hostname)) < 0) {
         if ((h = gethostbyname(hostname)) == 0)
            return err(-1, "%s: host not found\n", hostname);

         if (h->h_addrtype != AF_INET)
            return err(-1, "%s: not an internet host\n", hostname);

         bcopy((char *)*h->h_addr_list, (char *)&isock.sin_addr, h->h_length);
      }
      else
         isock.sin_addr.s_addr = iaddr;
      sock = (struct sockaddr *)&isock;
      len = sizeof(isock);
   }
#ifdef SO_SNDBUF
   if (bfactor > 1) {
      sendbuflen = sizeof(sendbufopt);
      if (getsockopt(s, SOL_SOCKET, SO_SNDBUF, &sendbufopt, &sendbuflen))
         perror("getsockopt[SNDBUF]");
      else {
         sendbufopt *= bfactor;
/*
warn("client_socket: changing sndbuf size to %d\n",sendbufopt);
*/
         if (setsockopt(s, SOL_SOCKET, SO_SNDBUF, &sendbufopt, sendbuflen))
            perror("getsockopt[SNDBUF]");
      }
   }
#endif
   if (connect(s, sock, len) < 0) {
      perror("connect");
      (void)close(s);
      return -1;
   }
#ifdef FIOSNBIO
   {
      int arg = 1;
      ioctl(s, FIOSNBIO, &arg);
   }
#else
   (void)fcntl(s, F_SETFL, FNDELAY);
#endif

   return s;
}
